//
// Gertbot: check UART is available
// If not try to fix it 
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int debug_mode = 1;

// Check for UART usage in cmd line and inittab
// 
// Return 1 on UART free
// Return 2 un UART busy
// return 3 on error
//
void check_uart_available(int *cmd_line_status,int *inittab_status)
{ FILE *fp;
  char string[1024];
  int line_numbers;
 
  fp = fopen("/boot/cmdline.txt","r");
  if (!fp)
    *cmd_line_status = 3;
  else
  { 
	  if (!fgets(string,1024,fp))
	    *cmd_line_status = 3;
	  else
		{ if (strstr(string,"ttyAMA0"))
		    *cmd_line_status = 2;
		  else
		    *cmd_line_status = 1;
		}
	  fclose(fp);
	}

  fp = fopen("/etc/inittab" ,"r");
  if (!fp)
    *inittab_status = 3;
  else
  {
    *inittab_status = 1;
    line_numbers = 0;
	  while (fgets(string,1024,fp))
	  { line_numbers++;
		  if (strstr(string,"ttyAMA0"))
		  { if (string[0]!='#')
		      *inittab_status=2;
		  }
		}
		fclose(fp);
		if (line_numbers==0)
		  *inittab_status = 3;
  }
} // check_uart_available

//
// Fix CMD LINE by removing ttyAMA0 from it
// This can only be done if this program is run in SUDO mode
//
// Return 1 in success
// Return 0 on failure 
//
// dwc_otg.lpm_enable=0 console=ttyAMA0,115200 kgdboc=ttyAMA0,115200 console=tty1 root=/dev/mmcblk0p2 rootfstype=ext4 elevator=deadline rootwait
int fix_cmdline()
{ FILE *fp;
  char string[1024]; // Hope this is enough (I know: sloppy programming!)
  char *tty_pos,*tty_start;
  int system_status;
 
  fp = fopen("/boot/cmdline.txt","r");
  if (!fp)
    return 0;
  if (!fgets(string,1024,fp))
  {
    fclose(fp);
    return 0;
  }

  // find ttyAMA0 in string and remove them 
  // this may leave extra spaces/tabs behind
  do {
    tty_pos = strstr(string,"ttyAMA0");
    if (tty_pos)
    { // find where tty argument starts
      while (tty_pos!=string && *tty_pos!=' ' && *tty_pos!='\t') 
        tty_pos--;
      if (*tty_pos==' ' || *tty_pos=='\t')
        tty_pos++;
      // tty_pos now points to start of tty argument
      tty_start = tty_pos;
      
      // Find where argument ends
      while (*tty_pos && *tty_pos!=' ' && *tty_pos!='\t')
        tty_pos++;
      // tty_pos now points to end of string or a space/tab
      
      // remove tty text
      while (*tty_pos)
        *tty_start++ = *tty_pos++;
      *tty_start = 0;        
    }
  } while (tty_pos);
  
  // make backup
  system_status = system("cp /boot/cmdline.txt /boot/cmdline.txt.org\n");
  if (system_status!=0)
  { // Probably started without 'sudo'
    // stop here an now
    return 0;
  }
  // Open file for writing
  // This whipes the old file 
  fp = fopen("/boot/cmdline.txt","w");
  if (!fp)
  { if (debug_mode)
      fprintf(stderr,"Could not open cmd for writing\n");
    return 0;
  }
  
  if (fputs(string,fp)<0)
  { // oops writing failed!
    if (debug_mode)
      fprintf(stderr,"Writing failedn\n");
    fclose(fp);
    // Try to restore old file
    system("cp /boot/cmdline.txt.org /boot/cmdline.txt\n");
    system("sync\n");
    return 0;
  }
  fclose(fp);
  return 1;
} // fix_cmdline

//
// Fix inittab by commenting out lines with ttyAMA0 
// This can only be done if this program is run in SUDO mode
//
// Return 1 in success
// Return 0 on failure 
//
int fix_inittab()
{ FILE *fpr,*fpw;
  char string[1024]; // Hope this is enough (I know: sloppy programming!)
  int line_number;
  int system_status;
 
  fpr = fopen("/etc/inittab","r");
  if (!fpr)
    return 0;
  fpw = fopen("/etc/inittab.new","w");
  if (!fpw)
  { fclose(fpr);
    return 0;
  }
  
  // Work through the file
  // comment out lines which contain ttyAMA0
  line_number = 0;
  while (fgets(string,1024,fpr))
  { line_number++;
    if (strstr(string,"ttyAMA0"))
    {
      if (string[0]!='#')
        fputc('#',fpw);
    }
    fputs(string,fpw);
  }
  fclose(fpr);
  fclose(fpw);
  
  // Now move the files around
  system_status = system("mv /etc/inittab /etc/inittab.org\n");
  system_status = system("mv /etc/inittab.new /etc/inittab\n");
  system_status = system("sync\n");
  
   
  return 1;
} // fix_cmdline

char *this_program;

void usage()
{ fprintf(stderr,"Usage: %s [s][c|cf|cfr]\n",this_program);
  fprintf(stderr,"%s is a program to check and free the uart setup on the Raspberry-Pi\n",
                  this_program);
  fprintf(stderr,"It checks if the uart is used at boot time by reading /boot/cmdline.txt\n");
  fprintf(stderr,"And checks if the uart is used as terminal by reading /etc/inittab\n");
  fprintf(stderr,"The program can also fix the state if it finds the uart is in use\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"%s accepts the following arguments:\n",this_program);
  fprintf(stderr,"   c    : Check uart available only\n");
  fprintf(stderr,"   Return status:\n");
  fprintf(stderr,"      0 : Uart is available for the user\n");
  fprintf(stderr,"      1 : Uart is in use by the system\n");
  fprintf(stderr,"      2 : Don't know: errors encounter during checking\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"   cf   : Checks and fixes if required\n");
  fprintf(stderr,"          (Only possible if the program is run in 'sudo' mode)\n");
  fprintf(stderr,"   Return status:\n");
  fprintf(stderr,"   0 : sucessfull\n");
  fprintf(stderr,"       0 : Uart is available for the user\n");
  fprintf(stderr,"       1 : Uart has been fixed but we need re-boot\n");
  fprintf(stderr,"       2 : Don't know: errors encounter during checking/correcting\\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"   cfr  : Checks and fixes if required, then asks for reboot\n");
  fprintf(stderr,"          (Only possible if the program is run in 'sudo' mode)\n");
  fprintf(stderr,"   Return status:\n");
  fprintf(stderr,"       0 : Uart is available for the user\n");
  fprintf(stderr,"       2 : Don't know: errors encounter during checking/correcting\n");
  fprintf(stderr,"Adding 's' in front of the arguments make the program silent\n");
  fprintf(stderr,"     except for rebooting. It never silently reboots!\n");
  exit(2);
} // usage

#define CHECK_ONLY         1
#define CHECK_FIX          2
#define CHECK_FIX_REBOOT   3

int main(int argc,char *argv[])
{ int action,silent;
  int cmd_line_status,inittab_status;
  int errors;
  char answer[32],*argv1;
  
  this_program = argv[0];
  if (argc==1)
    usage();
  argv1 = argv[1];
  if (*argv1=='s')
  { silent=1;
    argv1++;
  }
  else
    silent = 0;
  if (!strcmp(argv1,"c"))
    action = CHECK_ONLY;
  else  if (!strcmp(argv1,"cf"))
    action = CHECK_FIX;
  else  if (!strcmp(argv1,"cfr"))
    action = CHECK_FIX_REBOOT;
  else
    usage();
 
  check_uart_available(&cmd_line_status,&inittab_status);
  if (!silent)
  {  switch (cmd_line_status)
     {
     case 1 : printf("/boot/cmdline.txt seems OK\n");
              break;
     case 2 : printf("Looks like /boot/cmdline.txt needs fixing\n");
              break;
     case 3 : printf("Error checking /boot/cmdline.txt\n");
              break;
     }
     switch (inittab_status)
     {
     case 1 : printf("/etc/inittab seems OK\n");
              break;
     case 2 : printf("Looks like /etc/inittab needs fixing\n");
              break;
     case 3 : printf("Error checking /etc/inittab\n");
              break;
     }
  } // if not silent 
  
  // If everything seems ok exit with status 0 
  if (cmd_line_status==1 && inittab_status==1)
    return 0;
  
  // On errors with either file just stop
  if (cmd_line_status==3 || inittab_status==3)
    return 2; // Just exit now 
  
  // If only needed to check we are done now 
  if (action == CHECK_ONLY)
    return 1; 
  
  // We are in check and fix mode
  // Report what we are doing
  errors = 0;
  if (cmd_line_status==2)
  {
    if (!silent)
    { printf("Trying to fix /boot/cmdline.txt\n");
      fflush(stdout); // make sure message go out
    }
    if (!fix_cmdline())
    { if (!silent)
      { printf("Fix /boot/cmdline.txt failed\n");
        printf("Did you forget to run this command in 'sudo' mode??\n");
      }
      errors++;
    }
    else
      if (!silent)
        printf("Fix /boot/cmdline.txt seem OK\n");
  } // try fix cmd_line

  if (inittab_status==2)
  {
    if (!silent)
    { printf("Trying to fix /etc/inittab\n");
      fflush(stdout); // make sure message go out
    }

    if (!fix_inittab())
    { if (!silent)
      { printf("Fix /etc/inittab failed\n");
        printf("Did you forget to run this command in 'sudo' mode??\n");
      }
      errors++;
    }
    else
      if (!silent)
        printf("Fix /etc/inittab seem OK\n");
  } // fix inittab
  
  if (errors)
  { if (!silent)
      printf("Errors encountered during fixing: exiting\n");
    return 2;
  }

  
  
  if (action==CHECK_FIX)
  {
    if (!silent && (inittab_status==2 || cmd_line_status==2))
    {
      printf("*****\n");
      printf("*****  System seems fixed for uart use by user\n");
      printf("*****  But these changes do not take effect until you reboot!!!\n");
      printf("*****\n");
    }
    return 1;
  }

  // action must be CHECK_FIX_REBOOT
  // Use error channel for this text
  do {
    fprintf(stderr,"Uart seems fixed.\n");
    fprintf(stderr,"These changes only go into effect on reboot.\n");
    fprintf(stderr," ****************************\n");
    fprintf(stderr," *                          *\n");
    fprintf(stderr," *  REBOOT SYSTEM NOW Y/N?  *\n");
    fprintf(stderr," *                          *\n");
    fprintf(stderr," ****************************\n");
    fgets(answer,32,stdin);
  } while (answer[0]!='y' && answer[0]!='n' && 
           answer[0]!='Y' && answer[0]!='N');
  
  if (answer[0]=='y' || answer[0]=='Y')
  { 
    system("sync"); // yes, I am paranoid
    system("sudo reboot\n");
    // This program never finishes from here
  }
  
  return 1; // status: fixed but no reboot
} // main 

